<?php
namespace ServerlessFC;

use Psr\Http\Message\ServerRequestInterface;
use RingCentral\Psr7\Response;
use RingCentral\Psr7\Uri;

trait FcErrorTrait {
    function fc_exception_error_handler($errno, $errstr, $errfile, $errline) {
        $logger  = $GLOBALS['fcLogger'];
        $errInfo = array(
            "errorMessage" => $errstr,
            "errorType"    => \ServerlessFC\friendly_error_type($errno),
            "stackTrace"   => array(
                "file" => $errfile,
                "line" => $errline,
            ),
        );
        throw new \ErrorException($errstr, 0, $errno, $errfile, $errline);
    }

    public function set_fc_error_handler() {
        set_error_handler(array($this, 'fc_exception_error_handler'));
    }
}

trait FcLoadTrait {
    public function load($__fc_func_file) {

        //loaded mods can be reloaded
        include_once $__fc_func_file;

        //copy defined global vars to global scope
        $defined = get_defined_vars();

        unset($defined["__fc_func_file"]);

        $GLOBALS += $defined;
    }

}

// NormalCARequest include initialize and invoke request
class NormalCARequest {

    use FcErrorTrait;
    use FcLoadTrait;

    private $func_name;

    private $file_name;

    public function __construct($file, $func) {
        $this->set_fc_error_handler();

        $this->file_name = $file;
        $this->func_name = $func;

        $this->load($this->file_name);
    }

    public function invoke($event, $context) {

        $fname = $this->func_name;

        if (!function_exists($fname)) {
            throw new \Exception("function " . $fname . " not found");
        }
        $ret = $fname($event, $context);
        if (!is_string($ret)) {
            $ret = json_encode($ret, JSON_UNESCAPED_UNICODE);
        }
        return $ret;

    }

    public function initialize($context) {

        $fname = $this->func_name;

        if (!function_exists($fname)) {
            throw new \Exception("function " . $fname . " not found");
        }
        $ret = $fname($context);
        if (!is_string($ret)) {
            $ret = json_encode($ret, JSON_UNESCAPED_UNICODE);
        }
        return $ret;

    }
}

class HttpInvokeHandler {

    use FcErrorTrait;
    use FcLoadTrait;

    private $func_name;

    private $file_name;

    public function __construct($file, $func) {
        $this->set_fc_error_handler();

        $this->file_name = $file;
        $this->func_name = $func;

        $this->load($this->file_name);
    }

    public function handle(ServerRequestInterface $request, array $context): Response{
        $request  = $this->prepare_request($request, $context);
        $fname    = $this->func_name;
        $response = $fname($request, $context);
        assert($response instanceof Response);

        $resp_headers = $response->getHeaders();
        $resp_params  = array(
            'status'     => $response->getStatusCode(),
            'headersMap' => $resp_headers,
        );

        $params_str       = json_encode($resp_params);
        $resp_http_params = base64_encode($params_str);

        // construct fc headers
        foreach ($resp_headers as $key => $value) {
            $response = $response->withoutHeader($key);
        }

        $response = $response->withHeader(FC_HTTP_PARAMS, $resp_http_params)
            ->withHeader('Content-Type', 'application/octet-stream')
            ->withStatus(200);

        return $response;
    }

    private function prepare_request(ServerRequestInterface $request, $context) {
        $fname = $this->func_name;

        if (!function_exists($fname)) {
            throw new \Exception("function " . $fname . " not found");
        }

        $httpParams = $request->getHeaderLine(FC_HTTP_PARAMS);

        if ($httpParams == '') {
            throw new \Exception($fname . "httpParams is necessary!");
        }

        $httpParamsJsonStr = base64_decode($httpParams);
        $requestArr        = json_decode($httpParamsJsonStr, $assoc = true);

        // remove headers from ea(except content-type, content-length)
        $allHeaders = $request->getHeaders();
        foreach ($allHeaders as $h => $v) {
            if (in_array(strtoupper($h), ['CONTENT-TYPE', 'CONTENT-LENGTH'])) {
                continue;
            }
            $request = $request->withoutHeader($h);
        }

        //use client header  to replace header
        $headersMap = array_key_exists('headersMap', $requestArr) ? $requestArr['headersMap'] : [];
        foreach ($headersMap as $h => $v) {
            $request = $request->withHeader($h, $v);
        }

        $f = function ($n) {
            if (is_array($n) && count($n) == 1) {
                return $n[0];
            } else {
                return $n;
            }
        };

        $queriesMap = array_key_exists('queriesMap', $requestArr) ? $requestArr['queriesMap'] : [];
        if ($queriesMap) {
            $queries = array_map($f, $queriesMap);
            $request = $request->withQueryParams($queries);
        }

        $method     = array_key_exists('method', $requestArr) ? $requestArr['method'] : '';
        $path       = array_key_exists('path', $requestArr) ? $requestArr['path'] : '';
        $requestURI = array_key_exists('requestURI', $requestArr) ? $requestArr['requestURI'] : '';
        $clientIP   = array_key_exists('clientIP', $requestArr) ? $requestArr['clientIP'] : '';
        $request    = $request->withMethod($method);

        $format  = '%s.%s.fc.aliyuncs.com%s';
        $url     = sprintf($format, $context['accountId'], $context['region'], $requestURI);
        $fcUri   = new Uri($url);
        $request = $request->withUri($fcUri);

        $serverParams                = $request->getServerParams();
        $serverParams['REMOTE_ADDR'] = $clientIP;
        $request                     = $request->withServerParams($serverParams);

        $request = $request->withAttribute("clientIP", $clientIP)
            ->withAttribute("requestURI", $requestURI)
            ->withAttribute("path", $path);

        return $request;
    }
}
